package esmith::AD;

use strict;
use Net::LDAP;
use MIME::Base64();
use esmith::ConfigDB;

#This package is to aid in debugging.  Can remove
use Data::Dumper;

no warnings ('qw');

=head1 NAME

esmith::AD - Interface to Samba 4 Active Directory for 
             Koozali SME Server

=head1 SYNOPSIS

    use esmith::AD::User;
    use esmith::AD::Group;

    $ad_user = esmith::AD::User->new();
    @users = $ad_user->listUsers();

    $ad_group = esmith::AD::Group-new();
    @groups = $ad_group->listGroups();

    NOTE: Most methods in the class should not be run directly, but
          from a child class such as esmith::AD::User, or as noted 
          in the method specific documentation.

    
=head1 DESCRIPTION

This module provides an abstracted interface to Samba 4 Active Directory
for the Koozali SME Server.

Methods provided by this interface interact with Samba 4 Active Directory
using the Net::LDAP perl module.  Understanding of Net::LDAP syntax is
not required to utilize the primary functionality provided by this interface,
but the full suite of Net::LDAP methods is provided by this module should 
the developer wish to use them.

=cut

sub new {
    my ($class, %params) = @_;
    my $self = {};

    bless ($self, $class);
    return ($self);
}

=head2 Fully Abstraced Methods

For all intents and purposes, fully abstracted methods are the the preferred 
approach to interact with the Koozali Active Directory. Included in this API 
are child classes for managing the different object types in the Active Directory 
including users, groups, and DNS entries.  Methods provided in this parent 
class should generally not be called directly (unless noted in the method 
specific documentation), but from one of these child classes. This is because 
this API provides run-time binding of many of these parent methods to the 
object types they are presented with. See esmith::AD::User, esmith::AD::Group, 
esmith::AD:OU, and esmith::AD::DNS for more details.

Usage example:
       use esmith::AD::User;
       my $ad = esmith::AD->new;
       my $user_status = $ad->getUserStatus;


=head3 deleteAttr($name,$hashRef)  

This method deletes attribute values for an Active Directory object 
given the sAMAccount name for the object and a hash references of 
attribute/value pairs to delete.   This method does not delete the actual
attribute from the object, as attributes are built into the schema for all
objects, but just the value(s0.  To delete multiple values for multi-value
attributes, pass the values to delete as an array reference. 

Usage:
    
 
=cut 

sub deleteAttr {
    my ($self,$name,$attrRef) = @_;
   
    return 0 unless ($name && $attrRef && (ref($attrRef) eq 'HASH')); 
 
    #Don't allow this method from this class
    die $self->error("Must call this method on a esmith::AD child object\n")
      if (ref($self) eq 'esmith::AD');
    
    #Query Record and perform delete
    my ($record) = $self->queryObjects($name);
    die $self->error("Unable to query \"$name\" from Active Directory.\n.")
      unless ($record);   
 
    my $result = $self->deleteAD($record,$attrRef);
    die $self->error("Unable to values for \"$name\" record\n") 
      unless ($result);

    return (1);
}


=head3 getADRecord($name)

This method returns a hash reference of attribute/value pairs 
from the Active Directory for attributes supported by this method,
given an object name.  This method must be called from a child 
method or this code will return a runtime error. 

Usage:  $hashref = $ad->getADRecord($name);

=cut

sub getADRecord {
    my ($self,$name) = @_;
    
    die $self->error("Must call this method on a esmith::AD child object.\n")
      if (ref($self) eq 'esmith::AD');
    die $self->error("Object name not specified.\n") unless $name;
    
    #Query record from AD
    my ($record) = $self->queryObjects($name);
 
    #Verify that we found a record in the AD
    die $self->error("Unable to find \"$name\" in the Active Directory.\n") 
      unless ($record);

    #Build a hash of supported attributes and values 
    my @attributes = $self->attributes;
    
    my %adRecord = map {$_ => ($record->get_value($_))||''} @attributes;
   
    return (\%adRecord);
}


=head3 getName($dn)

This method returns an object sAMAccountName given an object DN. 

Usage:  $sAMAccountName= = $ad->getName(DN);

=cut

sub getName {
    my ($self,$dn) = @_;
     
    return (0) unless $dn;

    my @dnParts = split (/\,/,$dn);
    my ($cn,$name) = split (/=/,$dnParts[0]);
    
    return ($name); 
}


=head3 getSID($name)

This method retrieves the SID for a specified Active Directory 
object name, decrypts it, and then returns it.  This method must be 
called from a child method or this code will return a runtime error.

Usage: my $sid = $ad->getSID('username');

=cut

sub getSID {
    my ($self,$name) = @_;
    
    #Don't allow method from this class
    die $self->error("Must call this method on a esmith::AD child object.\n")
      if (ref($self) eq 'esmith::AD');
    die $self->error("Object name not specified.\n") unless $name;

    my ($record) = $self->queryObjects($name);

    die $self->error("Unable to find \"$name\" in the Active Directory.\n")
      unless ($record);
   
    my $encryptedSID = $record->get_value('objectSid');

    return($self->sidToString($encryptedSID));
}


=head3 getStatus($name)

This method returns the status, enabled or disabled, for various entries
in the Active Directory. This method must be called from a child
method or this code will return a runtime error.

Usage:  my $status=$ad->getStatus('username');

=cut

sub getStatus {
    my ($self,$name) = @_;
    
    #Don't allow this method from this class
    die $self->error("Must call this method on a esmith::AD child object.\n")
      if (ref($self) eq 'esmith::AD');
    die $self->error("Object name not specified.\n") unless $name;
    
    #Lookup entry in the active directory
    my ($record) = $self->queryObjects($name);
    
    die $self->error("Unable to find \"$name\" in the Active Directory.\n") 
      unless ($record);
    
    #Determine if Active Directory entry is enabled or disabled
    my $UAC = $record->get_value('userAccountControl');
   
    if ($UAC == 512 || $UAC == 66048) {return('enabled');}
    else {return ('disabled');}
}


=head3 listObjects($set)

This method returns a list of object names (sAMAccountName) from the Active 
Directory.  It will return one of the following subsets of AD objects:

   - all:     All searchable objects;
   - builtin: Objects defined during the AD provisioning;
   - koozali: Objects defined by this API;

Usage example 1: Get a list of koozali users
 
       use esmith::AD::User;
       my $ad = esmith::AD;User->new();
       my @kooazli_users = $ad->listObjects('koozali');

Usage example 2:  Get list of buildin groups

       use esmith::AD:Group;
       my $ad = esmith::AD:Group->new();
       my @builtin_groups = $ad->listObjects('builtin');

This method must be called from a child method or this code will return a runtime error.

=cut

sub listObjects 
{
    my ($self,$set) = @_;

    ##Must call from child object so we get runtime parameters
    die $self->error("Must call this method on a esmith::AD child object.\n")
      if (ref($self) eq 'esmith::AD');

    ## Determine what the subset of objects to search for 
    unless (grep {$set eq $_} qw(all builtin koozali)) 
    {
        $set = 'all';
    }

    #Run query
    my @records = $self->queryObjects($set); 

    #Return empty array even if there was an error with the query. 
    return () unless (@records);
 
    #successful query   
    my @objects = map { $_->get_value('sAMAccountName') } @records;
    return(@objects);
}


=head3 queryObjects($self,$setValue,$attribute,$attrValue,$base)

This is the primary query method for this API and returns Net::LDAP record
objects from the Active Directory in list context. Many of the query elements 
for an AD query are set via runtime parameters in the child class calling 
this method. If nothing is passed to this method, then this method returns 
a list of all user records in the Active Directory.

$setValue defines what records this method will search for, given default query
elements set in the child class.  Setting $value to the following will search 
for specific groups of user records in the AD:

 - all:  All user account records  in the active directory
 - builitin: Builtin user account records.  Example:  Administrator
 - koozali: All user account records defined in this UI. These are the typical
            user accounts setup in the server-manager for via the commandline.

Setting $setValue to a value other than those listed above will assume that this 
is the value of the record you are searching for.  If $attrValue is set, then this
value is always assumed to be the search value.  This allows for queries of sets 
(i.e., all, builtin, or koozali) by an attribute defined with the $attribute 
parameter.  This method will automatically runtime bind to the correct object type 
(i.e., users, groups, OUs, DNS). The default search attribute is the username 
(e.g., sAMAccountName). 

Searchs by AD attributes other than sAMAccountName need to be defined explicently 
in the child classes new() method. This prevents users from running queries that 
would result in AD errors. See the attribute hash in the new() method in child 
classes for details.   
       
       #Search for a user record by username
       use esmith::AD::User;
       my $ad = esmith::AD->new;
       my ($userRecord) = $ad->queryObjects('username');

       #Search all Koozali users
       user esmith::AD::User;
       my @koozaliUser = $ad->queryObjects('koozali');

=cut

sub queryObjects 
{
    my ($self,$setValue,$attribute,$attrValue,$base)=@_;
    
    #Error out if this method called from this class and not a child
    die $self->error("Must call this method on a esmith::AD child object.\n")
      if (ref($self) eq 'esmith::AD');

    #Check for changes to the runtime query parameters.
    $attribute = 'default' 
      unless ($attribute && defined($self->{'attribute'}->{$attribute}));
    $base = 'default' 
      unless ($base && defined($self->{'base'}->{$base}));
   
    #Assume we are searching for all user accounts unless the $value tells us
    #otherwise
    my $set = 'all';
    my $searchValue = '';

    #Look at $setValue and determine query set.  If $setValue is set to 
    #somthing other than all, builtin, koozali; then we'll assume this is 
    #the attribute value, unless $attrValue is also set.  $attrValue 
    #overrides a single setValue, if this is set. 
    if (grep {$setValue eq $_} qw(all builtin koozali)) 
    {
        $set = $setValue;
    }
    else 
    {
        $set = 'single';
        $searchValue = $setValue;
    }

    #Check for $attrValue. If it's set, we'll use that for the 
    #query value.  
    if (defined $attrValue) {$searchValue = $attrValue;}

    #Build Query Filter
    my $queryFilter = $self->{'base'}->{"$base"} .
                      $self->{'set'}->{"$set"};

    if ($set eq 'single' || defined $attrValue) 
    {
        $queryFilter .= "(" . 
                        $self->{'attribute'}->{$attribute} . 
                        "=" .
                        $searchValue .
                        ")";
    }

    #Return a list of record objects
    return ($self->runQuery($queryFilter));
}


=head3 setAttr($name,$attr,$value)

This method sets or modifies a single attribute/value pair for a 
specified object given the object sAMAccount Name and a attribute
value pair.  See setManyAttr for more details about updating object
attributes.

Usage:  
       $ad->setAttr('name','attribute','value')

=cut 

sub setAttr 
{
    my ($self,$name,$attr,$value) = @_;

    if ($self->setManyAttr($name,{$attr=>$value})) { return(1); }
    else { die->error("Unable to modify \"$attr\" for \"$name\".\n");}
}


=head3 setManyAttr($name,$hashRef)  

This method sets or modifies multiple attribute / value pairs for 
an Active Directory object sAMAccount. The method uses
runtime binding to determine which type of object is being updated.
Update attribute pairs need to be passed to the method as a hash 
reference. It should be noted that this method will only update 
attribute / value pairs supported by this class.  All other pairs will
be ignored.  Supported attribute/value paris are defined in the 
validData method. This method returns 1 if successful and will typically
die with appropriate error information if unsucessful. 

NOTE:  This method currently dies if data fails validation check.  
       This may need to be changed to provide better feedback to 
       server-panels in the future.
=cut 

sub setManyAttr {
    my ($self,$name,$attrRef) = @_;
   
    return 0 unless ($name && $attrRef); 
 
    #Don't allow this method from this class
    die $self->error("Must call this method on a esmith::AD child object\n")
      if (ref($self) eq 'esmith::AD');
    
    #Check to make sure we have enough data to perform an update
    die $self->error("No attributes set to modify for $name\n") 
      unless (ref($attrRef) eq 'HASH');
    
    #Validate that we support the attributes for update
    foreach my $key (keys %$attrRef) {
        unless ($self->validData->{$key}) {
            warn "Ignoring $key attribute for update -- Not supported\n";
            delete $attrRef->{$key};
        }
        if($attrRef->{$key} eq '') {
            warn "Ignoring $key attribute since it is blank\n";
            delete $attrRef->{$key};
        }
    }    

    #Query Record to update from AD
    my ($record) = $self->queryObjects($name);
    die $self->error("Unable to query \"$name\" from Active Directory.\n.")
      unless ($record);   
 
    my $result = $self->updateAD($record,$attrRef);
    die $self->error("Unable to update \"$name\" record\n") unless ($result);

    return (1);
}

=head3 setPosix($type,$name)

This method designates an Active Directory object as a Posix object given
and object sAMAccount name (e.g., username).  

=cut 

sub setPosix {
    my ($self,$name) = @_;

    #Don't allow this method from this class
    die $self->error("Must call this method on a esmith::AD child object.\n")
      if (ref($self) eq 'esmith::AD');
 
    #Pull record from AD
    my ($record) = $self->queryObjects($name);
    die $self->error("Unable to find \"$name\" in the Active Directory.\n") 
      unless ($record);
    
    ##Localize posix type array reference to shorten LDAP update syntax   
    my $typeRef = {'objectClass' => $self->{'type'}->{'posix'}};

    #Set the Posix objectClass
    my $result = $self->updateAD($record,$typeRef);
    die $self->error("Unable to update \"$name\" record.\n") unless ($result);

    return(1);
}


=head2 Partially Abstraced Methods

Methods providing a thin abstracted connection to Koozali Active Directory
using Net::LDAP.  These methods are designed to connect directly to the 
Koozali Active Directory and return Net::LDAP objects.  The developer should
generally not need to call these Partially Abstracted Methods directly, as 
this class includes Fully Abstracted Methods to more directly and specifically
access data in the Active Directly.  However, the developer may utilize these 
Partially Abstracted Methods for more fine grained interacting with the 
Active Directory using Net::LDAP.

=head3 connect()

This method returns a connection object to the Active Directory
using configuration parameters stored in the Koozali configuration
database and the currently provisioned Active Directory domain.

Domain administrator credentials needed to access the Active Directory
are taken from a password encrypted to the local filesystem in 
/etc/samba/AD.pw.  A method to decrypt this stashed password is 
provided by the esmith::AD::getADPass().

    Usage:
           $self = esmith::AD->new;
           $ad_connection_object = $self->connect;

=cut

sub connect {
    my $self = shift;
    
    #Pull Samba Domain and AD IP config db 
    my $domain = $self->domain;
  
    #Pull AD password from local storage
    my $ad_password = $self->getADPass;
    
    #Open a connection to the Active Director and return the connection
    #my $ad = Net::LDAP->new($sme_ad_host);
    my $ad = Net::LDAP->new('127.0.0.1'); 

    #Bind to the AD and return the connection object
    my $bind = $ad->bind($domain . "\\ad_admin", password=>$ad_password);

    if($bind->code) {
        die $self->error($bind->server_error);
    }
    
    return ($ad); 
}


=head3 deleteAD($record,$attRef)

This method will delete either attribute values or an entire record from 
the Active Directory, given a Net::LDAP::Entry object.  Passing a hash 
references of attribute/value pairs will delete the values for the attributes
specified in the hash reference.  For multi-value attributes, pass an array
reference of values for the attribute. See perldoc for 
Net::LDAP::Entry->delete() for more details.

This method does not check the contents of the hash reference, but relies upon
error reporting from Net::LDAP. 

Usage:    

=cut 

sub deleteAD {
    my ($self,$record,$attrRef) = @_;

    die $self->error("Record object not specified.\n")
      unless ($record =~ /Net::LDAP::Entry/); 
    
    #Bind to AD and update 
    my $ad = $self->connect();

    if((ref($attrRef) eq "HASH")) 
    {
        $record->delete(%$attrRef); #Delete local attribute values
    }
    else 
    { 
        $record->delete(); #Delete entire local record 
    }
   
    my $result = $record->update($ad); #update LDAP
    die $self->error($result->server_error) if($result->code());
 
    $ad->unbind;
    
    return(1);
}

 
=head3 runQuery($filter)

This method runs a query of the Active Directory given a query filter
formated as specified by Net::LDAP and returns a list of Net::LDAP::Entry
objects. It is not necessary to bind to the Active Directory before calling 
this meathod, as this method calls the connect() method itself.

Usage:
       use esmith::AD;
 
       my $ad = esmith::AD->new;
       my @records = $ad->runQuery($filter);

=cut 

sub runQuery {
    my ($self,$filter) = @_;

    die $self->error('Query filter not specified.') unless $filter;

    ##Bind to AD 
    my $ad = $self->connect();
  
    ##Get LDAP base DN for AD query -- we need to build it.
    my $baseDN = $self->baseDN;

    ##Query the Active Directory
    my $query=$ad->search(
                          base   => "$baseDN",
                          filter => "$filter"
                         );
    $ad->unbind;
 
    if($query->code) {die $self->error($query->server_error);}

    return ($query->entries);
}

=head3 updateAD($record,$attRef)

This method updates an Active Directory record, given a Net::LDAP::Entry
object and a hash reference of data to update.  This method does not check 
the contents of the record to update nor the hash reference prior to trying 
to update the Active Directory, other than schema checks done by the LDAP 
server. The method will return 1 if the update was successful and will die 
with a LDAP server error is there was a problem.

Usage:    
       my $update_result = $self->updateAD($record);

=cut 

sub updateAD {
    my ($self,$record,$attrRef) = @_;

    die $self->error("Record object to update not specified.\n")
      unless ($record =~ /Net::LDAP::Entry/); 
    die $self->error("Update hash reference not specified.\n")
      unless (ref($attrRef) eq "HASH"); 

    #Bind to AD and update 
    my $ad = $self->connect();

    $record->replace(%$attrRef); #Update local record
    my $result = $record->update($ad); #update LDAP
    
    if($result->code()) { die $self->error($result->server_error); }
    $ad->unbind;
    
    return(1);
}

##-----------------------------------------------------------
##HELPER METHODS 
##
##The methods in the following section are not documented 
##other than the note comments contained in this code, as they 
##are meant to support other methods contained in this class.
##The developer may use these methods, but should examine the
##code closely to achieve desired results.
##-----------------------------------------------------------

#------------------------------------------------------------
#Method: baseDN()
#  This method builds the Base DN for LDAP query of the 
#  Active Directory.  The Base DN is build from the 
#  primary machine domain name, which is considered the 
#  realm by Samba.
#     
#-----------------------------------------------------------
sub baseDN {
    my ($self)=@_;

    my $realm = $self->realm;
    my @realmParts =  split (/\./, $realm);

    my @dnParts = map {'DC=' . $_} @realmParts;
    my $base = join (',' , @dnParts);
    
    return ($base);
}

#------------------------------------------------------------
#Method: error()
#-----------------------------------------------------------
sub error {
    my ($self,$error)=@_;
    
    $error = '' unless $error;
 
    return ((caller(1))[3] . " error: " . $error);
}

#------------------------------------------------------------
#Method:  domain()
#   This method returns the Samba domain, which is stored
#   in the configuration dbase as Workgroup
#------------------------------------------------------------
sub domain {
    my $self = shift;
    
    my $cdb = esmith::ConfigDB->open_ro;
    my $domain = $cdb->get_prop('smb','Workgroup') ||
        die $self->error("Unable to determine the Workgroup from ConfigDB\n");
    $cdb->close;

    return ($domain);
}

#------------------------------------------------------------
#Method: getObjectDN()
#  This method builds and object DN for user and group 
#  objects in the Active Directory.  
#-----------------------------------------------------------
sub getObjectDN {
    my ($self,$name)=@_;

    return 0 unless $name;

    my $base = $self->baseDN;
    
    return('CN=' . $name . ',' . 'CN=Users,' . $base); 
}

#------------------------------------------------------------
#Method:  realm()
#   This method returns the realm which is the SME 
#   domain name stored in the configuration dbase
#   This method is not meant to be used external 
#   to this class.
#------------------------------------------------------------
sub realm {
    my $self = shift;

    my $cdb = esmith::ConfigDB->open_ro;
    my $realm = $cdb->get_prop('smb','realm');
    return ($realm) if defined $realm;

    my $domainName = $cdb->get_value('DomainName') ||
      die $self->error("Unable to determine Domain Name from ConfigDB\n");
    $cdb->close;
    
    my $ad_domain = $self->domain();
    
    return($ad_domain . '.' . $domainName);
}

#------------------------------------------------------------
#Method:  getADPass()
#   This method returns the locally stashed and encrypted
#   AD password.
#-----------------------------------------------------------
sub getADPass {
    my $self = shift;

    open( ADPW,"</etc/samba/AD.pw") ||
      die $self->error("Unable to read AD stashed password " .
                       "from local storage.\n");
    my $encrypted = <ADPW> || '';
    chomp $encrypted;
    close ADPW;
    return (MIME::Base64::decode($encrypted));
}

#------------------------------------------------------------
#Method:  hostIP()
#   This method returns the Samba domain from the smb
#   configuration dbase. This method is not meant to be
#   used external to this class.
#-----------------------------------------------------------
sub hostIP {
    my $self=shift;
    
    my $cdb = esmith::ConfigDB->open_ro;
    my $hostip = $cdb->get('LocalIP')->value ||
      die $self->error("Unable to determine AD server IP from ConfigDB.\n");
    $cdb->close;
    
    return($hostip);
}

#------------------------------------------------------------
#Method:  sidToString
#
#   This method decryptes the Samba SIDs and returns a
#   clear text version of the SID
#-----------------------------------------------------------
sub sidToString {
    my ($self, $sid) = @_;

    return undef
        unless unpack("C", substr($sid, 0, 1)) == 1;

    ##FIX ME:  This is a check for the SID length, but it doesn't work.
    ##Fortunately this function seems to be working fine without this
    ##check, but it would be nice to figure out why this is failing
    #return undef
    #    unless length($sid) == 8 + 4 * unpack("C", substr($sid, 1, 1));
    
    my $sid_str = "S-1-";
    
    $sid_str .= (unpack("C", substr($sid, 7, 1)) +
                (unpack("C", substr($sid, 6, 1)) << 8) +
                (unpack("C", substr($sid, 5, 1)) << 16) +
                (unpack("C", substr($sid, 4, 1)) << 24));
    
    for my $loop (0 .. unpack("C", substr($sid, 1, 1)) - 1) {
        $sid_str .= "-" . unpack("I", substr($sid, 4 * $loop + 8, 4));
    }

    return $sid_str;
}

=head1 COPYRIGHT

Copyright (c) 2016 Koozali Foundation, Inc.
This program is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.

=head1 AUTHOR

Greg Zartman, Koozali Foundation <gzartman@koozali.org>

=cut

=head1 SEE ALSO

Net::LDAP

=cut

1;
